#![deny(warnings)]
use js_sys::Date;
use sauron::{
    prelude::*,
    web_sys::MouseEvent,
};

pub enum Msg {
    Click,
    DoubleClick,
    Clock,
    ChangeName(String),
    ChangeBiography(String),
    ChangeThought(String),
}

pub struct App {
    click_count: u32,
    double_clicks: u32,
    date: Date,
    name: String,
    biography: String,
    thought: Option<String>,
}

impl App {
    pub fn new(click_count: u32) -> App {
        App {
            click_count,
            double_clicks: 0,
            date: Date::new_0(),
            name: String::new(),
            biography: String::new(),
            thought: None,
        }
    }
}

impl Component<Msg> for App {
    fn update(&mut self, msg: Msg) -> Cmd<Self, Msg> {
        match msg {
            Msg::Click => {
                self.click_count += 1;
            }
            Msg::DoubleClick => {
                self.double_clicks += 1;
            }
            Msg::Clock => {
                self.date = Date::new_0();
            }
            Msg::ChangeName(name) => {
                self.name = name;
            }
            Msg::ChangeBiography(bio) => {
                self.biography = bio;
            }
            Msg::ChangeThought(thought) => {
                if thought.len() > 0 {
                    self.thought = Some(thought);
                } else {
                    self.thought = None;
                }
            }
        }
        Cmd::none()
    }

    fn view(&self) -> Node<Msg> {
        let date_str: String = self
            .date
            .to_locale_string("en-GB", &JsValue::undefined())
            .into();
        div(
            vec![class("some-class"), id("some-id"), attr("data-id", 1)],
            vec![
                div(
                    vec![id("current-time")],
                    vec![text!("Today is {}", date_str)],
                ),
                div(
                    vec![],
                    vec![
                        text("Your name is: "),
                        input(
                            vec![
                                r#type("text"),
                                on_input(|event: InputEvent| {
                                    Msg::ChangeName(event.value.to_string())
                                }),
                                placeholder("John Smith"),
                            ],
                            vec![],
                        ),
                        button(
                            vec![on_click(|event: MouseEvent| {
                                trace!(
                                    "Clicked at ({},{})",
                                    event.x(),
                                    event.y()
                                );
                                Msg::Click
                            })],
                            vec![text!("Click me!")],
                        ),
                        button(
                            vec![on_dblclick(|event: MouseEvent| {
                                trace!(
                                    "Double clicked at ({},{})",
                                    event.x(),
                                    event.y()
                                );
                                Msg::DoubleClick
                            })],
                            vec![text!("DoubleClicks {}", self.double_clicks)],
                        ),
                    ],
                ),
                p(
                    vec![],
                    vec![
                        text!("Hello {}!", self.name),
                        if self.click_count > 0 {
                            text!(
                                ", You've clicked on that button for {} time{}",
                                self.click_count,
                                if self.click_count > 1 { "s" } else { "" }
                            )
                        } else {
                            span(vec![], vec![])
                        },
                    ],
                ),
                div(
                    vec![],
                    vec![
                        p(
                            vec![],
                            vec![text!("Tell us something about yourself:")],
                        ),
                        div(
                            vec![],
                            vec![textarea(
                                vec![
                                    rows(10),
                                    cols(80),
                                    on_input(|event: InputEvent| {
                                        Msg::ChangeBiography(
                                            event.value.to_string(),
                                        )
                                    }),
                                    placeholder("I'm a..."),
                                ],
                                vec![],
                            )],
                        ),
                        p(vec![], vec![text!("{}", self.biography)]),
                    ],
                ),
                div(
                    vec![],
                    vec![
                        text("What are you thinking right now?"),
                        input(
                            vec![
                                type_("text"),
                                on_change(|event: InputEvent| {
                                    Msg::ChangeThought(event.value.to_string())
                                }),
                                placeholder("Elephants..."),
                            ],
                            vec![],
                        ),
                        if let Some(thought) = &self.thought {
                            text(format!("Hmmn {}... Interesting.", thought))
                        } else {
                            span(vec![], vec![])
                        },
                    ],
                ),
            ],
        )
    }
}
